#ifndef eCharBuffer_H
#define eCharBuffer_H

using namespace std;

class eCharBuffer
{
private:
	/// Amount of space currently allocated.
	int reserved;
	/// Amount of data currently stored.
	int stored;
	/// Amount of space free
	int space_free;
	
	/// Pointer to buffer.
	unsigned char *buffer;
	
	/// Flag to allow buffer growth when adding more data
	bool CanGrow;
	
	/// pointer to the beginning of buffer
	unsigned char *ptr_pbase;
	/// pointer to the beginning of readable data
	unsigned char *ptr_gptr;
	/// pointer to the end of readable data, the byte just after the data
	unsigned char *ptr_egptr;
	/// pointer to the end of the buffer, the last byte in buffer
	unsigned char *ptr_eptr;
	
public:
	/// Default constructor.
	eCharBuffer(bool GrowFlag=true);
	/// constructor.
	eCharBuffer(int n, bool GrowFlag=true);
	//eCharBuffer(){};
	/// Destructor.
	~eCharBuffer();

	/** @return The amount of storage currently allocated */
	inline int allocated(void) const { return reserved; }

	/// return the amount of actual data available
	inline int used(void) const { return stored; }
	
	// return the amount of free space
	inline int available(void) const { return space_free; }
	
	// set whether I can grow my size
	inline void AllowGrow(bool flag) { CanGrow = flag; }
	
	/// pointer to the beginning of buffer
	const unsigned char * pbase(void) const { return ptr_pbase; }
	/// pointer to the beginning of readable data
	const unsigned char * gptr(void) const { return ptr_gptr; }
	/// pointer to the end of readable data
	const unsigned char * egptr(void) const { return ptr_egptr; }
	/// pointer to the end of the buffer
	const unsigned char * eptr(void) const { return ptr_eptr; }	
	
	/// add n more chars to the buffer 
	int push(char *, int n);
	
	/// remove n chars from the buffer, return free space
	int scrub(int n);
	
	/// remove all data from the buffer, returns free space
	int flush(void); 
	
	/// assign a new size
	bool NewSize(int new_size);
};

#endif
