#include <cstdio>
#include <string>
#include <iostream>
#include "eCharBuffer.h"

using namespace std;

eCharBuffer::eCharBuffer(bool GrowFlag) :
	reserved(2),
	buffer(new unsigned char[2]),
	CanGrow(GrowFlag)
{
	ptr_pbase = buffer;
	ptr_gptr  = buffer;
	ptr_egptr = buffer;
	ptr_eptr  = buffer + reserved - 1;
	stored = 0;
	space_free = reserved - stored;
}


eCharBuffer::eCharBuffer(int n, bool GrowFlag) :
	reserved(n),
	buffer(new unsigned char[reserved]),
	CanGrow(GrowFlag)
{
	ptr_pbase = buffer;
	ptr_gptr  = buffer;
	ptr_egptr = buffer;
	ptr_eptr  = buffer + reserved - 1;
	stored = 0;
	space_free = reserved - stored;
}


eCharBuffer::~eCharBuffer()
{
	ptr_pbase = NULL;
	ptr_gptr  = NULL;
	ptr_egptr = NULL;
	ptr_eptr  = NULL;
	delete [] buffer;
}

// Push more data into the buffer
int eCharBuffer::push(char * src, int n)
{
	// nothing to do
	if(n < 1) return(0);
	
	int space_free_end;

	space_free_end = ptr_eptr - ptr_egptr + 1;

	if ( (space_free_end) >= n)    // this is cheap
	{
//	cerr << "Request to push "<<n<<" onto "<<stored<<":"<<reserved<<" stuffing"<<endl;

		memcpy(ptr_egptr, src, n); // store new data, locations do not overlap
		ptr_egptr += n;           //update pointers
		stored += n;			   //update total stored
		space_free = reserved - stored;		
		return(n);                 //how much additional data stored
	}
	else if( space_free >= n) // this cost more
	{
//	cerr << "Request to push "<<n<<" onto "<<stored<<":"<<reserved<<" shuffling"<<endl;

		memmove(ptr_pbase, ptr_gptr, stored); // memmove, locations overlap
		ptr_gptr = ptr_pbase;          //update pointers
		ptr_egptr = ptr_gptr + stored;     //update pointers
		memcpy(ptr_egptr, src, n); // store new data, locations do not overlap
		ptr_egptr += n;                //update pointers
		stored += n;			       //update total stored
		space_free = reserved - stored;
		return(n);	                   //how much additional data stored
	}
	else if(CanGrow)    // this is expensive
	{
//	cerr << "Request to push "<<n<<" onto "<<stored<<":"<<reserved<<" growing"<<endl;

		int new_space_needed;
		unsigned char *tmp_ptr;
		
		// space needed and then some extra
		//total new      = (additional need) + current + extra
		new_space_needed = (n - space_free) + reserved + 1024;
//		cerr << "CharBuffer: New space needed is "<<new_space_needed <<endl;	
		tmp_ptr = new unsigned char[new_space_needed];
		if (tmp_ptr == NULL)
		{
			cerr << "eCharBuffer::push failed to allocate additional memory";
			return(0); //how much was stored
		}
		reserved = new_space_needed;           //update total reserved
		memcpy(tmp_ptr, ptr_gptr, stored); // copy old data, no overlap
		delete [] buffer;
		buffer = tmp_ptr;
		ptr_pbase = buffer;
		ptr_gptr = buffer;
		ptr_egptr = buffer + stored;
		ptr_eptr = buffer + reserved - 1;
		cerr << "Move complete, now add new data"<<endl;
		memcpy(ptr_egptr, src, n); // store new data, no overlap
		ptr_egptr += n;                //update pointers
		stored += n;			       //update total stored
		space_free = reserved - stored;
		return(n);	                   //how much additional data stored
	}
	else   // you told me not to grow, this is all i can offer now
	{
		cerr << "DANGER! DANGER! CharBuffer could not grow, didn't store"<< endl;
		return(0); // how much additional data stored
	}
	
	return(-1); // shouldn't get here
}

int eCharBuffer::scrub(int n)
{
//	cerr << "Request to scrub "<<n<<" from "<<stored<<":"<<reserved<<endl;
	if (n <= 0) return(space_free);	
	else if ( n >= stored) // wipe it all
	{
		ptr_gptr = ptr_pbase;  //update pointers
		ptr_egptr = ptr_pbase; //update pointers
		stored = 0;            //update total stored
		space_free = reserved - stored;
		return(space_free);
	}
	else
	{
		ptr_gptr += n;   //update pointers
		stored -= n;     //update total stored
		space_free = reserved - stored;
		return(space_free);
	}

}

int eCharBuffer::flush(void)
{
//	cerr << "Request to flush "<<stored<<":"<<reserved<<endl;
	ptr_gptr = ptr_pbase;  //update pointers
	ptr_egptr = ptr_pbase; //update pointers
	stored = 0;            //update total stored 
	space_free = reserved - stored;
	return(reserved);
}

bool eCharBuffer::NewSize(int new_size)
{
//	cerr << "Request to resize to "<<new_size<<" from "<<reserved<<endl;

	if ( new_size > reserved) //need additional space
	{
		// actually allocate more space and move data to beginning
	
		unsigned char *tmp_ptr;
		
		tmp_ptr = new unsigned char[new_size];
		if (tmp_ptr == NULL)
		{
			cerr << "eCharBuffer::NewSize failed to allocate additional memory";
			return(0); //how much was stored
		}
		reserved = new_size;               //update total reserved
		memcpy(tmp_ptr, ptr_gptr, stored); // copy old data, no overlap
		delete [] buffer;
		buffer = tmp_ptr;
		ptr_pbase = buffer;
		ptr_gptr = buffer;
		ptr_egptr = buffer + stored;
		ptr_eptr = buffer + reserved - 1;
		space_free = reserved - stored;
		return(true);	                   //how much we added	
	}
	else if ( new_size == reserved) //no additional space needed
	{
		return(true);
	}
	else if ( new_size >= stored) //freeing some unused space
	{
		// actually allocate more space and move data to beginning
	
		unsigned char *tmp_ptr;
		
		tmp_ptr = new unsigned char[new_size];
		if (tmp_ptr == NULL)
		{
			cerr << "eCharBuffer::NewSize failed to allocate additional memory";
			return(0); //how much was stored
		}
		reserved = new_size;               //update total reserved
		memcpy(tmp_ptr, ptr_gptr, stored); // copy old data, no overlap
		delete [] buffer;
		buffer = tmp_ptr;
		ptr_pbase = buffer;
		ptr_gptr = buffer;
		ptr_egptr = buffer + stored;
		ptr_eptr = buffer + reserved - 1;
		space_free = reserved - stored;
		return(true);	                   //how much we added			
	}
	else
	{	
		// current data won't fit, remove some and try again
		return(false);	
	}
}
